from pyrisco.common import GROUP_ID_TO_NAME, Partition as BasePartition

class Partition(BasePartition):
  """A representation of a Risco partition."""

  def __init__(self, api, raw):
    """Read partition from response."""
    self._api = api
    self._raw = raw

  async def disarm(self):
    return await self._api.disarm(self.id)

  async def arm(self):
    return await self._api.arm(self.id)

  async def partial_arm(self):
    return await self._api.partial_arm(self.id)

  async def group_arm(self, group):
    return await self._api.group_arm(self.id, group)

  @property
  def id(self):
    """Partition ID number."""
    return self._raw["id"]

  @property
  def disarmed(self):
    """Is the partition disarmed."""
    return self._raw["armedState"] == 1

  @property
  def partially_armed(self):
    """Is the partition partially-armed."""
    return self._raw["armedState"] == 2

  @property
  def armed(self):
    """Is the partition armed."""
    return self._raw["armedState"] == 3

  @property
  def triggered(self):
    """Is the partition triggered."""
    return self._raw["alarmState"] == 1

  @property
  def exit_timeout(self):
    """Time remaining till armed."""
    return self._raw["exitDelayTO"]

  @property
  def arming(self):
    """Is the partition arming."""
    return self.exit_timeout > 0

  @property
  def groups(self):
    """Group arming status."""
    if self._raw.get("groups") is None:
      return {}
    return {GROUP_ID_TO_NAME[g["id"]]: g["state"] == 3 for g in self._raw["groups"]}



from pyrisco.common import GROUP_ID_TO_NAME, Zone as BaseZone

class Zone(BaseZone):
  def __init__(self, panel, zone_id, status, zone_type, label, partitions, groups, tech):
    self._panel = panel
    self._id = zone_id
    self._status = status
    self._type = zone_type
    self._name = label.strip()
    self._partitions = partitions
    self._groups = int(groups, 16)
    self._tech = tech

  async def bypass(self, bypass):
    """Bypass or unbypass the zone."""
    return await self._panel.bypass_zone(self.id, bypass)

  @property
  def id(self):
      """Zone ID number."""
      return self._id

  @property
  def name(self):
      """Zone name."""
      return self._name

  @property
  def type(self):
      """Zone type."""
      return self._type

  @property
  def triggered(self):
      """Is the zone triggered."""
      return 'O' in self._status

  @property
  def alarmed(self):
      """Is the zone causing an alarm."""
      return 'a' in self._status

  @property
  def bypassed(self):
      """Is the zone bypassed."""
      return 'Y' in self._status

  @property
  def groups(self):
      """Groups the zone belongs to."""
      return [GROUP_ID_TO_NAME[i] for i in range(0,4) if ((2**i) & self._groups) > 0]

  @property
  def partitions(self):
      """partitions the zone belongs to."""
      ps = zip([int(p, 16) for p in self._partitions], range(0, len(self._partitions)))
      return [i*4 + p + 1 for c, i in ps for p in range(0,4) if ((2**p) & c) > 0]

  def update_status(self, status):
    self._status = status



import base64

CRC_ARRAY_BASE64 = 'WzAsNDkzNDUsNDk1MzcsMzIwLDQ5OTIxLDk2MCw2NDAsNDk3MjksNTA2ODksMTcyOCwxOTIwLDUxMDA5LDEyODAsNTA2MjUsNTAzMDUsMTA4OCw1MjIyNSwzMjY0LDM0NTYsNTI1NDUsMzg0MCw1MzE4NSw1Mjg2NSwzNjQ4LDI1NjAsNTE5MDUsNTIwOTcsMjg4MCw1MTQ1NywyNDk2LDIxNzYsNTEyNjUsNTUyOTcsNjMzNiw2NTI4LDU1NjE3LDY5MTIsNTYyNTcsNTU5MzcsNjcyMCw3NjgwLDU3MDI1LDU3MjE3LDgwMDAsNTY1NzcsNzYxNiw3Mjk2LDU2Mzg1LDUxMjAsNTQ0NjUsNTQ2NTcsNTQ0MCw1NTA0MSw2MDgwLDU3NjAsNTQ4NDksNTM3NjEsNDgwMCw0OTkyLDU0MDgxLDQzNTIsNTM2OTcsNTMzNzcsNDE2MCw2MTQ0MSwxMjQ4MCwxMjY3Miw2MTc2MSwxMzA1Niw2MjQwMSw2MjA4MSwxMjg2NCwxMzgyNCw2MzE2OSw2MzM2MSwxNDE0NCw2MjcyMSwxMzc2MCwxMzQ0MCw2MjUyOSwxNTM2MCw2NDcwNSw2NDg5NywxNTY4MCw2NTI4MSwxNjMyMCwxNjAwMCw2NTA4OSw2NDAwMSwxNTA0MCwxNTIzMiw2NDMyMSwxNDU5Miw2MzkzNyw2MzYxNywxNDQwMCwxMDI0MCw1OTU4NSw1OTc3NywxMDU2MCw2MDE2MSwxMTIwMCwxMDg4MCw1OTk2OSw2MDkyOSwxMTk2OCwxMjE2MCw2MTI0OSwxMTUyMCw2MDg2NSw2MDU0NSwxMTMyOCw1ODM2OSw5NDA4LDk2MDAsNTg2ODksOTk4NCw1OTMyOSw1OTAwOSw5NzkyLDg3MDQsNTgwNDksNTgyNDEsOTAyNCw1NzYwMSw4NjQwLDgzMjAsNTc0MDksNDA5NjEsMjQ3NjgsMjQ5NjAsNDEyODEsMjUzNDQsNDE5MjEsNDE2MDEsMjUxNTIsMjYxMTIsNDI2ODksNDI4ODEsMjY0MzIsNDIyNDEsMjYwNDgsMjU3MjgsNDIwNDksMjc2NDgsNDQyMjUsNDQ0MTcsMjc5NjgsNDQ4MDEsMjg2MDgsMjgyODgsNDQ2MDksNDM1MjEsMjczMjgsMjc1MjAsNDM4NDEsMjY4ODAsNDM0NTcsNDMxMzcsMjY2ODgsMzA3MjAsNDcyOTcsNDc0ODksMzEwNDAsNDc4NzMsMzE2ODAsMzEzNjAsNDc2ODEsNDg2NDEsMzI0NDgsMzI2NDAsNDg5NjEsMzIwMDAsNDg1NzcsNDgyNTcsMzE4MDgsNDYwODEsMjk4ODgsMzAwODAsNDY0MDEsMzA0NjQsNDcwNDEsNDY3MjEsMzAyNzIsMjkxODQsNDU3NjEsNDU5NTMsMjk1MDQsNDUzMTMsMjkxMjAsMjg4MDAsNDUxMjEsMjA0ODAsMzcwNTcsMzcyNDksMjA4MDAsMzc2MzMsMjE0NDAsMjExMjAsMzc0NDEsMzg0MDEsMjIyMDgsMjI0MDAsMzg3MjEsMjE3NjAsMzgzMzcsMzgwMTcsMjE1NjgsMzk5MzcsMjM3NDQsMjM5MzYsNDAyNTcsMjQzMjAsNDA4OTcsNDA1NzcsMjQxMjgsMjMwNDAsMzk2MTcsMzk4MDksMjMzNjAsMzkxNjksMjI5NzYsMjI2NTYsMzg5NzcsMzQ4MTcsMTg2MjQsMTg4MTYsMzUxMzcsMTkyMDAsMzU3NzcsMzU0NTcsMTkwMDgsMTk5NjgsMzY1NDUsMzY3MzcsMjAyODgsMzYwOTcsMTk5MDQsMTk1ODQsMzU5MDUsMTc0MDgsMzM5ODUsMzQxNzcsMTc3MjgsMzQ1NjEsMTgzNjgsMTgwNDgsMzQzNjksMzMyODEsMTcwODgsMTcyODAsMzM2MDEsMTY2NDAsMzMyMTcsMzI4OTcsMTY0NDhd'
ENCRYPTION_FLAG_INDEX = 1
ENCRYPTION_FLAG_VALUE = 17

START = b'\x02'
END = b'\x03'
DLE = b'\x10'
ESCAPED_START = DLE + START
ESCAPED_END = DLE + END
ESCAPED_DLE = DLE + DLE

def _is_encrypted(message):
    return message[ENCRYPTION_FLAG_INDEX] == ENCRYPTION_FLAG_VALUE


class RiscoCrypt:
  def __init__(self, encoding='utf-8'):
    self._pseudo_buffer = None
    self._crc_decoded = list(map(int, base64.b64decode(CRC_ARRAY_BASE64).decode("utf-8")[1:-1].split(',')))
    self.encrypted_panel = False
    self._encoding = encoding

  def set_panel_id(self, panel_id):
    self._pseudo_buffer = RiscoCrypt._create_pseudo_buffer(panel_id)

  def encode(self, cmd_id, command, force_crypt=False):
    encrypted = bytearray()
    encrypted.append(2)
    encrypt = force_crypt or self.encrypted_panel
    if encrypt:
      encrypted.append(17)

    full_cmd = f'{cmd_id:02d}{command}\x17'
    crc = self._get_crc(full_cmd)
    full_cmd += crc
    chars = self._encrypt_chars(full_cmd.encode(self._encoding), encrypt)
    encrypted += chars

    encrypted.append(3)
    return encrypted;

  def decode(self, chars):
    self.encrypted_panel = _is_encrypted(chars)
    decrypted_chars = self._decrypt_chars(chars)
    decrypted = decrypted_chars.decode(self._encoding)
    raw_command = decrypted[0:decrypted.index('\x17')+1]
    command, crc = decrypted.split('\x17')

    if command[0] in ['N','B']:
      cmd_id = null
      command_string = command
    else:
      cmd_id = int(command[:2])
      command_string = command[2:]
    
    return [cmd_id, command_string, self._valid_crc(raw_command, crc)]

  def _encrypt_chars(self, chars, encrypt):
    position = 0;
    if encrypt:
      chars = bytearray(map(self._encrypt_decrypt_char, chars, range(len(chars))))
    chars = chars.replace(DLE, ESCAPED_DLE)
    chars = chars.replace(START, ESCAPED_START)
    chars = chars.replace(END, ESCAPED_END)
      
    return chars

  def _decrypt_chars(self, chars):
    decrypt = _is_encrypted(chars)
    initial_index = 2 if decrypt else 1
    escaped = chars[initial_index:-1]
    escaped = escaped.replace(ESCAPED_DLE, DLE)
    escaped = escaped.replace(ESCAPED_START, START)
    escaped = escaped.replace(ESCAPED_END, END)
    
    if decrypt:
      return bytes(map(self._encrypt_decrypt_char, escaped, range(len(escaped))))
    else:
      return escaped

  def _encrypt_decrypt_char(self, char, position):
    return char ^ self._pseudo_buffer[position]

  def _create_pseudo_buffer(panel_id):
    buffer_length = 255
    pseudo_buffer = bytearray(buffer_length)
    if panel_id == 0:
      return pseudo_buffer
    pid = panel_id
    num_array = [2, 4, 16, 32768]
    for i in range(buffer_length):
      n1 = 0
      n2 = 0
      for n1 in range(4):
        if (pid & num_array[n1]) > 0:
          n2 ^= 1
      pid = pid << 1 | n2
      pseudo_buffer[i] = (pid & buffer_length)
    return pseudo_buffer

  def _valid_crc(self, command, crc):
    if len(crc) != 4:
      return False

    for char in crc:
      if ord(char) > 127:
        return False

    computed = self._get_crc(command)
    return computed == crc

  def _get_crc(self, command):
    crc_base = 65535
    byte_buffer = bytearray(command, self._encoding)

    for b in byte_buffer:
      crc_base = crc_base >> 8 ^ self._crc_decoded[crc_base & 255 ^ b]

    return f'{crc_base:04X}'



from .const import PANEL_TYPE, PANEL_MODEL, PANEL_FW, MAX_ZONES, MAX_PARTS, MAX_OUTPUTS

def _rw032_capabilities(firmware):
  return {
    PANEL_MODEL: 'Agility 4',
    MAX_ZONES: 32,
    MAX_PARTS: 3,
    MAX_OUTPUTS: 4,
  }

def _rw132_capabilities(firmware):
  return {
    PANEL_MODEL: 'Agility',
    MAX_ZONES: 36,
    MAX_PARTS: 3,
    MAX_OUTPUTS: 4,
  }

def _rw232_capabilities(firmware):
  return {
    PANEL_MODEL: 'WiComm',
    MAX_ZONES: 36,
    MAX_PARTS: 3,
    MAX_OUTPUTS: 4,
  }

def _rw332_capabilities(firmware):
  return {
    PANEL_MODEL: 'WiCommPro',
    MAX_ZONES: 36,
    MAX_PARTS: 3,
    MAX_OUTPUTS: 4,
  }

def _rw432_capabilities(firmware):
  max_zones = 32
  max_outputs = 14
  parts = firmware.split('.')
  if int(parts[0]) >= 3:
    max_zones = 50
    max_outputs = 32

  return {
    PANEL_MODEL: 'LightSys',
    MAX_ZONES: max_zones,
    MAX_PARTS: 4,
    MAX_OUTPUTS: max_outputs,
  }

def _rw512_capabilities(firmware):
  max_zones = 64
  parts = list(map(int, firmware.split('.')))
  if ((parts[0] > 1) or
  (parts[0] == 1 and parts[1] > 2) or
  (parts[0] == 1 and parts[1] == 2 and parts[2] > 0) or 
  (parts[0] == 1 and parts[1] == 2 and parts[2] == 0 and parts[3] >= 7)):
    max_zones = 128;

  return {
    PANEL_MODEL: 'ProsysPlus|GTPlus',
    MAX_ZONES: max_zones,
    MAX_PARTS: 32,
    MAX_OUTPUTS: 262,
  };

PANELS = {
  'RW032': _rw032_capabilities,
  'RW132': _rw132_capabilities,
  'RW232': _rw232_capabilities,
  'RW332': _rw332_capabilities,
  'RP432': _rw432_capabilities,
  'RP512': _rw512_capabilities
}

def panel_capabilities(panel_type, firmware):
  normalized = panel_type.split(":")[0]
  firmware = firmware.split(" ")[0]
  caps = PANELS[normalized](firmware)
  return {**caps, **{PANEL_TYPE: panel_type, PANEL_FW: firmware}}



from pyrisco.common import GROUP_ID_TO_NAME, Partition as BasePartition

class Partition(BasePartition):
  def __init__(self, panel, partition_id, label, status):
      """Read partition from response."""
      self._panel = panel
      self._id = partition_id
      self._status = status
      self._name = label.strip()

  async def disarm(self):
    """Disarm the partition."""
    return await self._panel.disarm(self.id)

  async def arm(self):
    """Arm the partition."""
    return await self._panel.arm(self.id)

  async def partial_arm(self):
    """Partially-arm the partition."""
    return await self._panel.partial_arm(self.id)

  async def group_arm(self, group):
    """Arm a group on the partition."""
    return await self._panel.group_arm(self.id, group)

  @property
  def id(self):
      """Partition ID number."""
      return self._id

  @property
  def name(self):
      """Partition name."""
      return self._name

  @property
  def disarmed(self):
      """Is the partition disarmed."""
      return not (self.armed or self.partially_armed)

  @property
  def partially_armed(self):
      """Is the partition partially-armed."""
      return 'H' in self._status

  @property
  def armed(self):
      """Is the partition armed."""
      return 'A' in self._status

  @property
  def triggered(self):
      """Is the partition triggered."""
      return 'a' in self._status

  @property
  def ready(self):
      """Is the partition ready."""
      return 'R' in self._status

  @property
  def arming(self):
      """Is the partition arming."""
      # return self.disarmed and not self.ready
      return False

  @property
  def groups(self):
      """Group arming status."""
      return {GROUP_ID_TO_NAME[g]: (str(g+1) in self._status) for g in range(0,4)}

  def update_status(self, status):
    self._status = status



from .partition import Partition
from .zone import Zone

class Alarm:
  """A representation of a Risco alarm system."""

  def __init__(self, api, raw):
    """Read alarm from response."""
    self._api = api
    self._raw = raw
    self._partitions = None
    self._zones = None

  @property
  def partitions(self):
    """Alarm partitions."""
    if self._partitions is None:
      self._partitions = {p["id"]: Partition(self._api, p) for p in self._raw["partitions"]}
    return self._partitions

  @property
  def zones(self):
    """Alarm zones."""
    if self._zones is None:
      self._zones = {z["zoneID"]: Zone(self._api, z) for z in self._raw["zones"]}
    return self._zones



import asyncio
import copy
from .const import PANEL_TYPE, PANEL_MODEL, PANEL_FW, MAX_ZONES, MAX_PARTS, MAX_OUTPUTS
from .panels import panel_capabilities
from .partition import Partition
from .zone import Zone
from .risco_socket import RiscoSocket
from pyrisco.common import OperationError, GROUP_ID_TO_NAME


class RiscoLocal:
  def __init__(self, host, port, code, **kwargs):
    self._rs = RiscoSocket(host, port, code, **kwargs)
    self._panel_capabilities = None
    self._listen_task = None
    self._zone_handlers = []
    self._partition_handlers = []
    self._error_handlers = []
    self._default_handlers = []
    self._event_handlers = []
    self._zones = None
    self._partitions = None
    self._id = None

  async def connect(self):
    await self._rs.connect()
    panel_type = await self._rs.send_result_command("PNLCNF")
    firmware = await self._rs.send_result_command("FSVER?")
    self._panel_capabilities = panel_capabilities(panel_type, firmware)
    self._id = int(await self._rs.send_result_command("PNLSERD"))
    self._zones = await self._init_zones()
    self._partitions = await self._init_partitions()
    self._listen_task = asyncio.create_task(self._listen(self._rs.queue))

  async def disconnect(self):
    await self._rs.disconnect()
    self._listen_task.cancel()
    self._listen_task = None

  def add_error_handler(self, handler):
    return RiscoLocal._add_handler(self._error_handlers, handler)

  def add_event_handler(self, handler):
    return RiscoLocal._add_handler(self._event_handlers, handler)

  def add_zone_handler(self, handler):
    return RiscoLocal._add_handler(self._zone_handlers, handler)

  def add_partition_handler(self, handler):
    return RiscoLocal._add_handler(self._partition_handlers, handler)

  def add_default_handler(self, handler):
    return RiscoLocal._add_handler(self._default_handlers, handler)

  @property
  def id(self):
    return self._id

  @property
  def zones(self):
    return self._zones

  @property
  def partitions(self):
    return self._partitions

  async def disarm(self, partition_id):
    """Disarm a partition."""
    return await self._rs.send_ack_command(f'DISARM={partition_id}')

  async def arm(self, partition_id):
    """Arm a partition."""
    return await self._rs.send_ack_command(f'ARM={partition_id}')

  async def partial_arm(self, partition_id):
    """Partially-arm a partition."""
    return await self._rs.send_ack_command(f'STAY={partition_id}')

  async def group_arm(self, partition_id, group):
    """Arm a specific group on a partition."""
    if isinstance(group, str):
        group = GROUP_ID_TO_NAME.index(group) + 1

    return await self._rs.send_ack_command(f'GARM*{group}={partition_id}')

  async def bypass_zone(self, zone_id, bypass):
    """Bypass or unbypass a zone."""
    if self.zones[zone_id].bypassed != bypass:
      await self._rs.send_ack_command(F'ZBYPAS={zone_id}')

  def _add_handler(handlers, handler):
    handlers.append(handler)
    def _remove():
      handlers.remove(handler)
    return _remove

  async def _init_partitions(self):
    return await self._get_objects(1, self._panel_capabilities[MAX_PARTS], self._create_partition)

  async def _init_zones(self):
    return await self._get_objects(1, self._panel_capabilities[MAX_ZONES], self._create_zone)

  async def _get_objects(self, min, max, func):
    ids = range(min, min+max)
    temp = await asyncio.gather(*[func(i) for i in ids])
    return { o.id: o for o in temp if o }

  async def _create_partition(self, partition_id):
    try:
      status = await self._rs.send_result_command(f'PSTT{partition_id}?')
      if not 'E' in status:
        return None

      label = await self._rs.send_result_command(f'PLBL{partition_id}?')
    except OperationError:
      return None
    return Partition(self, partition_id, label, status)

  async def _create_zone(self, zone_id):
    try:
      zone_type = int(await self._rs.send_result_command(f'ZTYPE*{zone_id}?'))
      if zone_type == 0:
        return None

      tech = await self._rs.send_result_command(f'ZLNKTYP{zone_id}?')
      if tech.strip() == 'N':
        return None

      status = await self._rs.send_result_command(f'ZSTT*{zone_id}?')
      if status.endswith('N'):
        return None

      label = await self._rs.send_result_command(f'ZLBL*{zone_id}?')
      partitions = await self._rs.send_result_command(f'ZPART&*{zone_id}?')
      groups = await self._rs.send_result_command(f'ZAREA&*{zone_id}?')
      return Zone(self, zone_id, status, zone_type, label, partitions, groups, tech)
    except OperationError:
      return None

  def _zone_status(self, zone_id, status):
    z = self._zones[zone_id]
    z.update_status(status)
    RiscoLocal._call_handlers(self._zone_handlers, zone_id, copy.copy(z))

  def _partition_status(self, partition_id, status):
    p = self._partitions[partition_id]
    p.update_status(status)
    RiscoLocal._call_handlers(self._partition_handlers, partition_id, copy.copy(p))

  def _default(self, command, result, *params):
    RiscoLocal._call_handlers(self._default_handlers, command, result, *params)

  def _event(self, event):
    RiscoLocal._call_handlers(self._event_handlers, event)

  def _error(self, error):
    RiscoLocal._call_handlers(self._error_handlers, error)

  def _call_handlers(handlers, *params):
    if len(handlers) > 0:
      async def _gather():
        await asyncio.gather(*[h(*params) for h in handlers])
      asyncio.create_task(_gather())

  async def _listen(self, queue):
    while True:
      item = await queue.get()
      if isinstance(item, Exception):
        self._error(item)
        continue

      if item.startswith('CLOCK'):
        # safe to ignore these
        continue

      if item.startswith("EVENT="):
        self._event(item[6:])
        continue

      command, result, *params = item.split("=")

      if command.startswith('ZSTT'):
        self._zone_status(int(command[4:]), result)
      elif command.startswith('PSTT'):
        self._partition_status(int(command[4:]), result)
      else:
        self._default(command, result, *params)
